/*
 * Nomad Jukebox filesystem for Linux.
 *
 * Copyright (C) 2001 Jrg Prante <joergprante@gmx.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/**
 *   The USB interface of the Nomad Jukebox file system
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/ctype.h>
#include <linux/usb.h>
#include <asm/errno.h>

#include "njbfs.h"
#include "njb_usb.h"
#include "nomad_usb.h"
#include "track.h"
#include "playlist.h"



/* forward decl's */
static int njb_usb_connect(struct nomad_usb_data *nomad);
static int njb_usb_disconnect(struct nomad_usb_data *nomad);
static int njb_usb_open(struct nomad_usb_data *nomad,
			struct njbfs_fattr *fattr);
static int njb_usb_close(struct nomad_usb_data *nomad,
			 struct njbfs_fattr *fattr);
static int njb_usb_write(struct nomad_usb_data *nomad, char *name,
			 u_int32_t offset, u_int32_t count, void *buffer);
static int njb_usb_read(struct nomad_usb_data *nomad, char *name,
			u_int32_t offset, u_int32_t count, void *buffer);
static int njb_usb_rename(struct nomad_usb_data *nomad,
			  struct njbfs_fattr *fattr);
static int njb_usb_delete(struct nomad_usb_data *nomad, char *file);
static int njb_usb_create(struct nomad_usb_data *nomad, char *file);
static int njb_usb_check(struct nomad_usb_data *njb);
static int njb_usb_capture(struct nomad_usb_data *njb);
static int njb_usb_handshake(struct nomad_usb_data *njb);
static int njb_usb_release(struct nomad_usb_data *njb);
static int njb_usb_control(struct nomad_usb_data *njb, unsigned int pipe,
			   int request, int requesttype, int value,
			   int index, void *data, int size, int retries);
static int njb_usb_verify_last_command(struct nomad_usb_data *njb);
static int njb_usb_update_library_counter(struct nomad_usb_data *njb);
static track_t *njb_usb_update_track_tag(struct nomad_usb_data *njb,
					 struct njbfs_fattr *fattr,
					 u_int32_t * trackid);
static int njb_usb_get_disk_usage(struct nomad_usb_data *njb,
				  u_int64_t * total, u_int64_t * free);

#ifdef DEBUG
void dump_data(unsigned char *data, int count);
#endif



/**
 *  Connect a Nomad Jukebox on the USB.
 *  This will be called when the device shall benmounted.
 *
 *  @param njb the USB structure of the NJB
 *  @param 0 if success, otherwise < 0
 */
static int njb_usb_connect(struct nomad_usb_data *njb)
{
	int status;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	dbg("connected to USB");

	return 0;
}


/**
 *  Disconnect a Nomad Jukebox on the USB.
 *  This will be called when the device shall be unmounted.
 *
 *  @param njb the USB structure of the NJB
 *  @param 0 if success, otherwise < 0
 */
static int njb_usb_disconnect(struct nomad_usb_data *njb)
{
	int status;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	dbg("disconnected from USB");

	return 0;
}


/**
 *  Open a Nomad Jukebox on the USB.
 *
 *  @param njb the USB structure of the NJB
 *  @param fattr the file attributes
 */
static int
njb_usb_open(struct nomad_usb_data *njb, struct njbfs_fattr *fattr)
{
	int status;
	track_t *track;
	u_int32_t trackid;
	u_int64_t total, free;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	if (njb->isopen) {
		err("open: too many open files");
		return -EMFILE;
	}

	if (!fattr) {
		err("open: fatal, no file attributes");
		return -EINVAL;
	}

	if (fattr->fileid == 0) {

		dbg("open: new file");

		if (fattr->size == 0) {
			err
			    ("open: sorry, no file size provided, can't continue.");
			return -EINVAL;
		}

		if ((fattr->title == NULL) || (fattr->artist == NULL)
		    || (fattr->codec == NULL)) {
			err("open: sorry, missing title/artist/codec");
			return -EINVAL;
		}

		/* capture the NJB */
		status = njb_usb_capture(njb);
		if (status < 0) {
			err("open: capture error, status = %d", status);
			return status;
		}

		/* handshake, checks NJB ID und library counter */
		status = njb_usb_handshake(njb);
		if (status < 0) {
			err("open: handshake error, status = %d", status);
			status = njb_usb_release(njb);
			return status;
		}
		dbg("open: handshake ok");

		/* Get disk usage */
		status = njb_usb_get_disk_usage(njb, &total, &free);
		if (status < 0) {
			err("open: get disk usage error, status = %d",
			    status);
			return -ENOSPC;
		}
		dbg("open: disk usage, total=%d, free=%d",
		    (u_int32_t) total, (u_int32_t) free);

		if (free < (u_int64_t) fattr->size) {
			err("open: no space left on device");
			return -ENOSPC;
		}
	/**
         * Build track, write track tag, and get NJB track id as a result
         */
		trackid = 0;
		track = njb_usb_update_track_tag(njb, fattr, &trackid);
		if (track == NULL) {
			err("open: error writing track tag header");
			status = njb_usb_release(njb);
			return -EIO;
		}

		fattr->track = track;	/* save track */
		fattr->fileid = trackid;	/* for requesting the track tag later */

		return status;
	} else if (fattr->fileid > 0) {

		/* Open for read track */

		/* before we can handshake, capture the NJB */
		status = njb_usb_capture(njb);
		if (status < 0) {
			err("open: capture error, status = %d", status);
			return status;
		}

		status = njb_usb_handshake(njb);
		if (status < 0) {
			err("open: handshake error, status = %d", status);
			status = njb_usb_release(njb);
			return status;
		}

		/*
		 * Request track
		 */
		status = njb_usb_control(njb,
					 usb_sndctrlpipe(njb->nomad_dev,
							 NJB_ENDPOINT_CONTROL),
					 NJB_CMD_REQUEST_TRACK,
					 UT_WRITE | USB_TYPE_VENDOR |
					 USB_RECIP_OTHER, 1, 0,
					 &fattr->fileid, 4, 3);
		if (status < 0) {
			err
			    ("open: 'request track' error, status = %d, track id = %d",
			     status, fattr->fileid);
			status = njb_usb_release(njb);
			return status;
		}

		dbg("open: 'request track' command ok, track id = %d",
		    fattr->fileid);

		status = njb_usb_verify_last_command(njb);
		if (status < 0)
			err("open: verify error, status = %d", status);

	}

	return status;
}


/**
 *   Close NJB on USB.
 */
static int
njb_usb_close(struct nomad_usb_data *njb, struct njbfs_fattr *fattr)
{
	int status;
	unsigned char njbstate;
	int retries;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	if (!fattr) {
		err("close: no file attributes");
		return -EINVAL;
	}

	/* close */
	retries = 3;
	while (retries > 0) {
		status = njb_usb_control(njb,
					 usb_rcvctrlpipe(njb->nomad_dev,
							 NJB_ENDPOINT_CONTROL),
					 NJB_CMD_TRANSFER_COMPLETE,
					 UT_READ | USB_TYPE_VENDOR |
					 USB_RECIP_OTHER, 0, 0, &njbstate,
					 1, 3);
		if (status < 0) {
			err("close: can't send close command, status=%d",
			    status);
			wait_ms(250);
			retries--;
		}
		if (njbstate == NJB_OK) {
			break;
		}
		if (njbstate == NJB_DEVICE_BUSY) {
			wait_ms(250);
			retries--;
		}
	}


	if (fattr->fileid == 0) {

		/* update library counter */
		dbg("close: updating lib counter");
		status = njb_usb_update_library_counter(njb);
		if (status < 0) {
			err
			    ("close: updating lib count error, status = %d, ",
			     status);
			return -EPROTO;
		}
	}

	njb->isopen = 0;

	status = njb_usb_release(njb);
	dbg("close: jukebox closed");

	return status;
}


/**
 *  Read block from USB NJB
 */
static int
njb_usb_read(struct nomad_usb_data *njb, char *name,
	     u_int32_t offset, u_int32_t count, void *buffer)
{
	int status;
	unsigned char *data;
	u_int16_t lsw, msw;
	unsigned int actual_length;
	unsigned char param[8];
	int size = IBUF_SIZE + 68;
	unsigned char *dest;
	int i;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	if ((data = (unsigned char *) kmalloc(size, GFP_KERNEL)) == NULL) {
		err("read: couldn't kmalloc data");
		return -ENOMEM;
	}
	memset(data, 0, size);


	msw = MSW(offset);
	lsw = LSW(offset);
	memcpy(&param[0], &lsw, 2);
	memcpy(&param[2], &msw, 2);
	msw = MSW(count);
	lsw = LSW(count);
	memcpy(&param[4], &lsw, 2);
	memcpy(&param[6], &msw, 2);

	status = njb_usb_control(njb,
				 usb_sndctrlpipe(njb->nomad_dev,
						 NJB_ENDPOINT_CONTROL),
				 NJB_CMD_RECEIVE_FILE_BLOCK,
				 UT_WRITE | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, 0, 0, &param, 8, 3);
	if (status < 0) {
		err("read: njb control error");
		/* let's run into bulk timeout and do a soft landing. */
	}

	/*  Receive bulk read */
	status = usb_bulk_msg(njb->nomad_dev,
			      usb_rcvbulkpipe(njb->nomad_dev,
					      NJB_ENDPOINT_BULK), data,
			      size, &actual_length, USB_BULK_TIMEOUT);
	if (status < 0) {
		err
		    ("read: bulk read error, actual_length = %d, status = %d",
		     actual_length, status);
	}

	if (data[0] != NJB_OK) {
		err("read: fatal, NJB state = %d", (int) data[0]);
	}

	dest = (unsigned char *) buffer;
	for (i = 0; i < actual_length - 68; i++) {
		dest[i] = data[i + 68];
	}

	kfree(data);

	return actual_length - 68;
}


/**
 * Write block to USB NJB
 *
 * @param njb the NJB USB structure
 * @param name the file name
 * @param offset the offset of the block
 * @param count the number of bytes to write
 * @param buffer the buffer address
 * @return 0 if success, otherwise < 0
 */
static int
njb_usb_write(struct nomad_usb_data *njb, char *name,
	      u_int32_t offset, u_int32_t count, void *buffer)
{
	unsigned char njbstate;
	int retries;
	int status;
	u_int32_t msw, lsw;
	ssize_t actual_length;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	lsw = LSW(count);
	msw = MSW(count);

	/* extra loop because NJB can be busy when we are writing too fast */
	retries = 3;
	while (retries > 0) {

		status = njb_usb_control(njb,
					 usb_rcvctrlpipe(njb->nomad_dev,
							 NJB_ENDPOINT_CONTROL),
					 NJB_CMD_SEND_FILE_BLOCK,
					 UT_READ | USB_TYPE_VENDOR |
					 USB_RECIP_OTHER, lsw, msw,
					 &njbstate, 1, 3);
		if (status < 0) {
			err("write: njb control error, status = %d",
			    status);
			return -EIO;
		}

		if (njbstate == NJB_OK)
			break;
		if (njbstate == NJB_DEVICE_BUSY) {
			wait_ms(100);
			retries--;
		} else {
			err("write: NJB state = %d", (int) njbstate);
			return -EAGAIN;
		}
	}

	/*  Send data with bulk write */
	status = usb_bulk_msg(njb->nomad_dev,
			      usb_sndbulkpipe(njb->nomad_dev,
					      NJB_ENDPOINT_BULK), buffer,
			      count, &actual_length, USB_BULK_TIMEOUT);
	if (status < 0) {
		err
		    ("write: bulk write error, actual_length = %d, status = %d",
		     actual_length, status);
	}
	if (actual_length < count) {
		err
		    ("write: bulk short write, actual_length = %d, count = %d",
		     actual_length, count);
	}

	/* verify write */
	status = njb_usb_verify_last_command(njb);
	if (status < 0) {
		err("write: can't verify last command, status = %d",
		    status);
	}

	return actual_length;

}

/**
 *  Renaming a file on the NJB
 *
 *  @param fattr the info how to rename
 */

static int
njb_usb_rename(struct nomad_usb_data *njb, struct njbfs_fattr *fattr)
{
	track_t *track;

	dbg("rename: replacing track tag id = %d", fattr->fileid);

	track = njb_usb_update_track_tag(njb, fattr, &fattr->fileid);
	if (track == NULL) {
		err("rename: can't replace track tag for id=%d",
		    fattr->fileid);
		return -EIO;
	}
	dbg("rename: replace track tag ok");

	return 0;
}

/**
 *  Delete a file from the NJB
 */
static int njb_usb_delete(struct nomad_usb_data *nomad, char *file)
{
	info("TODO: usb delete");
	return 0;
}


/**
 *  Create a file on the NJB
 */
static int njb_usb_create(struct nomad_usb_data *njb, char *file)
{
	info("TODO: create");
	return 0;
}


/* -----------------------  NJB functions -------------------------- */

/**
 *  Check the internal state of the NJB connection.
 *  The NJB USB structure and the device should be plugged in.
 *  These are sanity checks. It is never assumed that a NJB
 *  structure is missing, this function makes it only harder to
 *  bypass the internal NJB connection management.
 *
 * @param njb the NJB USB structure
 * @return 0 if success, otherwise < 0
 */
int njb_usb_check(struct nomad_usb_data *njb)
{
	if (!njb) {
		err("check: driver not installed");
		return -ENXIO;
	}
	if (!(njb->nomad_dev)) {
		err("check: jukebox not connected");
		return -ENODEV;
	}
	return 0;
}

/**
 * Send NJB control message. On the USB, all communications are
 * initiated by the host through the control pipe. A request packet
 *  is sent, followed by a data phase for that request, if there is
 *  data associated with the request. The setup and data phases are
 * done on the control pipe, with the setup packet indicating the
 * direction flow for the data phase. Additional data may be sent
 * after the control message via the bulk data pipes in response to a
 * request.
 *
 * @param njb the NJB USB structure
 * @param pipe the control pipe
 * @param request the request
 * @param requesttype the type of the request
 * @param value a value for the request
 * @param index an index for the request
 * @param data the data buffer for the request (not NULL)
 * @param size the size of the data buffer
 * @param retries how many times the control message shall be retried
 */
static int
njb_usb_control(struct nomad_usb_data *njb, unsigned int pipe,
		int request, int requesttype, int value, int index,
		void *data, int size, int retries)
{
	int status;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	down(&(njb->lock));
	while (retries) {
		status =
		    usb_control_msg(njb->nomad_dev, pipe, request,
				    requesttype, value, index, data, size,
				    USB_CONTROL_TIMEOUT);
		if (status == -ETIMEDOUT) {
			dbg("control: got a timeout, retrying...");
			retries--;
		} else if (status < 0) {
			err("control: error, result = %d",
			    le32_to_cpu(status));
			retries = 0;
		} else {
			retries = 0;
		}
	}
	up(&(njb->lock));

	return 0;
}

/**
 * Capture the NJB on the USB
 * This command is used to allocate the device if there are
 * multiple hosts which have concurrent access to the NJB.
 * The device is locked until a release command is issued.
 *
 * @param njb the NJB USB structure
 * @return 0 if success, otherwise < 0
 *
 */
static int njb_usb_capture(struct nomad_usb_data *njb)
{
	int status;
	unsigned char njbstate;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	if (!njb->captured) {
		status = njb_usb_control(njb,
					 usb_rcvctrlpipe(njb->nomad_dev,
							 NJB_ENDPOINT_CONTROL),
					 NJB_CMD_CAPTURE_NJB,
					 UT_READ | USB_TYPE_VENDOR |
					 USB_RECIP_OTHER, 0, 0, &njbstate,
					 1, 3);
		if (status < 0)
			err("capture: error, status = %d", status);
		if (njbstate != NJB_OK)
			err("capture: NJB state = %d", (int) njbstate);
		else
			njb->captured = 1;
	} else {
		dbg("capture: device already captured");
	}

	return status;

}


/**
 * Release the NJB on the USB
 * This command is used to release the device after capturing.
 *
 * @param njb the NJB USB structure
 * @return 0 if success, otherwise < 0
 */
static int njb_usb_release(struct nomad_usb_data *njb)
{
	int status;
	unsigned char njbstate;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	if (njb->captured) {
		status = njb_usb_control(njb,
					 usb_rcvctrlpipe(njb->nomad_dev,
							 NJB_ENDPOINT_CONTROL),
					 NJB_CMD_RELEASE_NJB,
					 UT_READ | USB_TYPE_VENDOR |
					 USB_RECIP_OTHER, 0, 0, &njbstate,
					 1, 3);
		if (njbstate != NJB_OK)
			err("release: NJB state = %d", (int) njbstate);
		njb->captured = 0;
	}

	return status;
}

/**
 * Verify if the last NJB command can be executed.
 *
 * @param njb the NJB USB structure
 * @return 0 if success, otherwise < 0
 */
static int njb_usb_verify_last_command(struct nomad_usb_data *njb)
{
	int status;
	unsigned char njbstate;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	status = njb_usb_control(njb,
				 usb_rcvctrlpipe(njb->nomad_dev,
						 NJB_ENDPOINT_CONTROL),
				 NJB_CMD_VERIFY_LAST_CMD,
				 UT_READ | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, 0, 0, &njbstate, 1, 3);
	if (njbstate != NJB_OK)
		err("verify: NJB state = %d", (int) njbstate);

	return status;
}


/**
 * Ping the NJB and get some device info
 *
 * @param njb the NJB USB structure
 * @return 0 if success, otherwise < 0
 */
static int njb_usb_ping(struct nomad_usb_data *njb, njbid_t * njbid)
{
	unsigned char data[58];
	unsigned int actual_length;
	unsigned char dummy;
	int status;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	status = njb_usb_control(njb,
				 usb_sndctrlpipe(njb->nomad_dev,
						 NJB_ENDPOINT_CONTROL),
				 NJB_CMD_PING,
				 UT_WRITE | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, 0, 0, &dummy, 0, 3);
	if (status < 0) {
		err("ping: njb control error");
	}

	/*  receive bulk read */
	status = usb_bulk_msg(njb->nomad_dev,
			      usb_rcvbulkpipe(njb->nomad_dev,
					      NJB_ENDPOINT_BULK), data, 58,
			      &actual_length, USB_BULK_TIMEOUT);
	if (status < 0) {
		err("ping: bulk read error");
		return -EIO;
	} else if (actual_length < 58) {
		err("ping: read interrupted, actual_length = %d",
		    actual_length);
		return -EIO;
	}

	if (data[0] != NJB_OK) {
		err("ping: NJB state = %d", (int) data[0]);
		return -EIO;
	}

	/*  unfold data into njbid */
	memcpy(&njbid->id, &data[1], 16);
	njbid->fwMinor = (u_int8_t) data[19];
	njbid->fwMajor = (u_int8_t) data[20];
	memcpy(&njbid->productName, &data[25], 32);
	njbid->power = (u_int8_t) data[57];

	return 0;
}


/**
 * Set the library counter of the NJB on the USB
 *
 * @param njb the NJB USB structure
 * @return 0 if success, otherwise < 0
 */
static int
njb_usb_set_library_counter(struct nomad_usb_data *njb, u_int64_t count)
{
	int status;
	unsigned char data[8];
	u_int32_t msdw, lsdw;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	msdw = MSL(count);
	lsdw = LSL(count);

	memset(data, 0, 8);
	memcpy(data, &msdw, 4);
	memcpy(&data[4], &lsdw, 4);

	dbg("njb_usb_set_library_counter: %d %d", lsdw, msdw);

	return njb_usb_control(njb,
			       usb_sndctrlpipe(njb->nomad_dev,
					       NJB_ENDPOINT_CONTROL),
			       NJB_CMD_SET_LIBRARY_COUNTER,
			       UT_WRITE | USB_TYPE_VENDOR |
			       USB_RECIP_OTHER, 0, 0, &data, 8, 3);

}


/**
 * Get the library counter of the NJB on the USB
 *
 * @param njb the NJB USB structure
 * @return 0 if success, otherwise < 0
 */
static int
njb_usb_get_library_counter(struct nomad_usb_data *njb,
			    njblibctr_t * lcount)
{
	int status;
	unsigned char data[25];
	u_int32_t msdw, lsdw;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	memset(lcount, 0, sizeof(njblibctr_t));

	status = njb_usb_control(njb,
				 usb_rcvctrlpipe(njb->nomad_dev,
						 NJB_ENDPOINT_CONTROL),
				 NJB_CMD_GET_LIBRARY_COUNTER,
				 UT_READ | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, 0, 0, &data, 25, 3);

	if (data[0] != NJB_OK) {
		err("get_library_counter: NJB state = %d", (int) data[0]);
		return -EIO;
	}

	memcpy(&lcount->id, &data[1], 16);
	memcpy(&lsdw, &data[21], 4);
	memcpy(&msdw, &data[17], 4);

	lcount->count = ((u_int64_t) msdw << 32) | lsdw;

	dbg("get_library_counter: lsl=%d, msl=%d",
	    LSL(lcount->count), MSL(lcount->count));

	return 0;
}

/**
 *  Perform NJB handshake. Get library counter and ID, try to set values,
 *  and check if values can be set.
 *
 * @param njb the NJB USB structure
 * @return 0 if success, otherwise < 0
 */
static int njb_usb_handshake(struct nomad_usb_data *njb)
{
	int status;
	njblibctr_t lcount, lcount_new, lcount_old;
	njbid_t njbid;
	unsigned char *cp;
	int i;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	if (njb_usb_ping(njb, &njbid) < 0)
		return -ENODEV;

	memcpy(njb->id, njbid.id, 16);
	cp = njb->idstring;
	for (i = 0; i < 16; i++) {
		sprintf(cp, "%02x", njb->id[i]);
		cp += 2;
	}
	njb->idstring[32] = 0;

	if (njb_usb_get_library_counter(njb, &lcount) < 0)
		return -EIO;

	if (memcmp(njb->id, lcount.id, 16)) {
		err("handshake: bad NJB ID");
		return -EPROTO;
	}

	lcount_new = lcount;
	lcount_old = lcount;

	lcount_new.count++;

	if (njb_usb_set_library_counter(njb, lcount_new.count) < 0)
		return -EPROTO;

	if (njb_usb_verify_last_command(njb) < 0)
		return -EPROTO;

	if (njb_usb_get_library_counter(njb, &lcount) < 0)
		return -EPROTO;

	if (memcmp(njb->id, lcount.id, 16)) {
		err("handshake: bad NJB ID");
		return -EPROTO;
	}

	if (lcount.count != lcount_new.count) {
		err("handshake: bad counter");
		return -EPROTO;
	}

	if (njb_usb_set_library_counter(njb, lcount_old.count) < 0)
		return -EPROTO;

	if (njb_usb_verify_last_command(njb) < 0)
		return -EPROTO;

	if (njb_usb_get_library_counter(njb, &lcount) < 0)
		return -EPROTO;

	if (memcmp(njb->id, lcount.id, 16)) {
		err("handshake: bad NJB ID");
		return -EPROTO;
	}

	if (lcount.count != lcount_old.count) {
		err("handshake: bad counter");
		return -EPROTO;
	}

	njb->libcount = lcount.count;

	return 0;
}


/**
 * Update NJB library counter
 *
 * @param njb the NJB USB structure
 * @return 0 if success, otherwise < 0
 */
static int njb_usb_update_library_counter(struct nomad_usb_data *njb)
{
	int status;
	njblibctr_t lcount;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	if (njb->session_updated)
		return 0;

	if (njb_usb_get_library_counter(njb, &lcount) < 0)
		return -EPROTO;

	if (memcmp(njb->id, lcount.id, 16)) {
		err("update_library_counter: bad NJB ID");
		return -EPROTO;
	}

	lcount.count++;

	if (njb_usb_set_library_counter(njb, lcount.count) < 0)
		return -EPROTO;

	if (njb_usb_verify_last_command(njb) < 0)
		return -EPROTO;

	njb->session_updated = 1;
	njb->libcount++;

	return 0;
}


/**
 *  Get a track tag header from the NJB
 *
 *  @param njb the njb to get the playlist from
 *  @param header the header to be filled with the result
 *  @param cmd can be NJB_GET_FIRST_TRACK_TAG_HEADER or
 *           NJB_GET_NEXT_TRACK_TAG_HEADER
 *  @return 0 if everything is ok or <0 if there was an error
 */
static int
njb_usb_get_track_tag_header(struct nomad_usb_data *njb,
			     njbttaghdr_t * header, int cmd)
{
	int status;
	unsigned char data[9];


	if ((status = njb_usb_check(njb)) < 0)
		return status;

	status = njb_usb_control(njb,
				 usb_rcvctrlpipe(njb->nomad_dev,
						 NJB_ENDPOINT_CONTROL),
				 cmd,
				 UT_READ | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, 0, 0, &data, 9, 3);
	if (status < 0) {
		err("get_track_tag_header: usb control error, status = %d",
		    status);
		return EIO;
	}

	if (data[0] == NJB_TRACK_NOT_FOUND) {
		/* silent error, as we reach end of list */
		return -ENOENT;
	} else if (data[0] != NJB_OK) {
		err("get_track_tag_header: bad NJB state = %02x",
		    (int) data[0]);
		return -EAGAIN;
	}

	memcpy(&header->trackid, &data[1], 4);
	memcpy(&header->size, &data[5], 4);

	/*
	   ben:
	   my NJB seems to get the tag header 
	   size wrong fairly consistantly, but the header size
	   never seems to exceed about 400.  
	 */
	header->size = 800;
	return 0;
}



inline static int
njb_usb_get_first_track_tag_header(struct nomad_usb_data *njb,
				   njbttaghdr_t * header)
{
	return njb_usb_get_track_tag_header(njb, header,
					    NJB_CMD_GET_FIRST_TRACK_TAG_HEADER);
}

inline static int
njb_usb_get_next_track_tag_header(struct nomad_usb_data *njb,
				  njbttaghdr_t * header)
{
	return njb_usb_get_track_tag_header(njb, header,
					    NJB_CMD_GET_NEXT_TRACK_TAG_HEADER);
}


/**
 *  Get track tag. This request retrieves a specific track tag from the NJB.
 *  As with the "Get Playlist" request, the track ID is obtained from a
 *  previously "Get Track Tag Header" request.
 */
static track_t *njb_usb_get_track_tag(struct nomad_usb_data *njb,
				      njbttaghdr_t * header)
{
	int status;
	u_int16_t msw, lsw;
	void *data;
	track_t *track;
	unsigned char *dp;
	unsigned char dummy;
	int actual_length = 0;

	if ((status = njb_usb_check(njb)) < 0)
		return NULL;

	if (!header) {
		err("get_track_tag: fatal, no header");
		return NULL;
	}

	if ((data = kmalloc(header->size + 5, GFP_KERNEL)) == NULL) {
		err
		    ("get_track_tag: couldn't kmalloc track tag of size: %d",
		     header->size);
		return NULL;
	}

	lsw = LSW(header->trackid);
	msw = MSW(header->trackid);

	/*  send a get track command */
	status = njb_usb_control(njb,
				 usb_sndctrlpipe(njb->nomad_dev,
						 NJB_ENDPOINT_CONTROL),
				 NJB_CMD_GET_TRACK_TAG,
				 UT_WRITE | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, msw, lsw, &dummy, 0, 3);

	if (status < 0) {
		err("get_track_tag: njb control error");
		return NULL;
	}

	/*  receive track tag by usb bulk read */
	status = usb_bulk_msg(njb->nomad_dev,
			      usb_rcvbulkpipe(njb->nomad_dev,
					      NJB_ENDPOINT_BULK), data,
			      header->size + 5, &actual_length,
			      USB_BULK_TIMEOUT);

	if (status < 0) {
		err("get_track_tag: bulk read error");
		kfree(data);
		return NULL;
	} else if (actual_length < header->size) {
		header->size = actual_length;
	}

	/* the track tag is in NJB format, so we convert it to our own */
	dp = (unsigned char *) data;

	track = track_unpack(&dp[5], header->size);
	if (track == NULL) {
		err("get_track_tag: couldn't unpack track");
		kfree(data);
		return NULL;
	}
	/* copy header into track */
	track->header.trackid = header->trackid;
	track->header.size = header->size;

	kfree(data);
	return track;
}


/**
 *   Get disk usage
 *
 *  @param njb the njb to get the disk usage from
 *  @param total result variable for the total bytes result
 *  @param free result variable for the free bytes result
 *  @return status
 */
static int
njb_usb_get_disk_usage(struct nomad_usb_data *njb,
		       u_int64_t * total, u_int64_t * free)
{
	int status;
	unsigned char data[17];
	u_int32_t msdw, lsdw;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	status = njb_usb_control(njb,
				 usb_rcvctrlpipe(njb->nomad_dev,
						 NJB_ENDPOINT_CONTROL),
				 NJB_CMD_GET_DISK_USAGE,
				 UT_READ | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, 0, 0, &data, 17, 3);
	if (status < 0) {
		err("get_disk_usage: usb control error, status = %d",
		    status);
		return -EIO;
	}

	if (data[0]) {
		err("get_disk_usage: bad NJB state = %d", data[0]);
		return -EAGAIN;
	}

	memcpy(&msdw, &data[1], 4);
	memcpy(&lsdw, &data[5], 4);
	*total = ((u_int64_t) msdw << 32) | lsdw;

	memcpy(&msdw, &data[9], 4);
	memcpy(&lsdw, &data[13], 4);
	*free = ((u_int64_t) msdw << 32) | lsdw;

	return 0;
}

/**
 *  Send track tag. This command writes a new track tag to the NJB.
 *  @param njb the USB NJB
 *  @param tag the new track tag
 *  @return status
 */

static int
njb_usb_new_track_tag(struct nomad_usb_data *njb, void *tag,
		      njbttaghdr_t * header)
{
	int status;
	ssize_t actual_length;
	unsigned char data[5];

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	dbg("write_track: command 'send track tag' of size %d",
	    header->size);
	status =
	    njb_usb_control(njb,
			    usb_sndctrlpipe(njb->nomad_dev,
					    NJB_ENDPOINT_CONTROL),
			    NJB_CMD_SEND_TRACK_TAG,
			    UT_WRITE | USB_TYPE_VENDOR | USB_RECIP_OTHER,
			    0, 0, &header->size, 4, 3);
	if (status < 0) {
		err("write_track: usb control error, status = %d", status);
		return status;
	}

	/* Bulk transfer of tag */

	dbg("write_track: tag bulk transfer");
	status = usb_bulk_msg(njb->nomad_dev,
			      usb_sndbulkpipe(njb->nomad_dev,
					      NJB_ENDPOINT_BULK), tag,
			      header->size, &actual_length,
			      USB_BULK_TIMEOUT);
	if (status < 0) {
		err("write_track_tag: bulk write error");
		return status;
	} else if (actual_length < header->size) {
		err
		    ("write_track_tag: short bulk write, actual_length = %d",
		     actual_length);
		return -EIO;
	}
	dbg("write_track_tag: tag sent, actual_length = %d",
	    actual_length);

	/*  Read answer from NJB, one status byte, four track id bytes */
	memset(data, 0, 5);
	status = usb_bulk_msg(njb->nomad_dev,
			      usb_rcvbulkpipe(njb->nomad_dev,
					      NJB_ENDPOINT_BULK), data, 5,
			      &actual_length, USB_BULK_TIMEOUT);
	if (status < 0) {
		err("write_track_tag: bulk read error, status = %d",
		    status);
		return status;
	} else if (actual_length < 5) {
		err("write_track_tag: short bulk read, actual_length = %d",
		    actual_length);
		return -EIO;
	}

	if (data[0] != NJB_OK) {
		err("write_track_tag: invalid data, NJB code = %d",
		    data[0]);
		return -EIO;
	}

	/* Set track id, return track */
	memcpy(&header->trackid, &data[1], 4);
	dbg("write_track_tag: got new track id = %d", header->trackid);

	return 0;

}


/**
 *  Replace track tag. This command replaces a specific track tag on the NJB.
 *
 *  @param njb
 *  @param tag the new track tag
 *  @return status
 */
static int
njb_usb_replace_track_tag(struct nomad_usb_data *njb, void *tag,
			  njbttaghdr_t * header)
{
	int status;
	u_int16_t msw, lsw;
	ssize_t actual_length;
	unsigned char *data;
	unsigned char dummy;

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	data = (unsigned char *) kmalloc(header->size + 4, GFP_KERNEL);
	if (data == NULL) {
		err("replace_track_tag_header: no memory");
		return -ENOMEM;
	}

	lsw = LSW(header->size);
	msw = MSW(header->size);

	/* setup phase */
	status = njb_usb_control(njb,
				 usb_rcvctrlpipe(njb->nomad_dev,
						 NJB_ENDPOINT_CONTROL),
				 NJB_CMD_REPLACE_TRACK_TAG,
				 UT_WRITE | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, msw, lsw, &dummy, 0, 3);
	if (status < 0) {
		err("replace_track_tag: usb control error, status = %d",
		    status);
		return -EIO;
	}

	memcpy(&data[0], &header->trackid, 4);
	memcpy(&data[4], tag, header->size);

	/* send tag */
	status = usb_bulk_msg(njb->nomad_dev,
			      usb_sndbulkpipe(njb->nomad_dev,
					      NJB_ENDPOINT_BULK), data,
			      header->size + 4, &actual_length,
			      USB_BULK_TIMEOUT);
	if (status < 0) {
		err("replace_track_tag: bulk write error");
		return -EIO;
	} else if (actual_length < header->size + 4) {
		err
		    ("replace_track_tag: short bulk write, actual_length = %d",
		     actual_length);
		return -EIO;
	}

	return 0;
}


/**
 * Update track tag. We don't care if track exist or not. If track exist,
 * replace the old track tag with the new one given by the file attributes.
 *
 * @param njb the USB NJB device
 * @param fattr the file attributes with the new values for the track
 * @param trackid pointer to track id, 0 = new, > 0 id of tag to be replaced
 * @return a valid new track, or NULL if an error occured
 */
static track_t *njb_usb_update_track_tag(struct nomad_usb_data *njb,
					 struct njbfs_fattr *fattr,
					 u_int32_t * trackid)
{
	int status;
	track_t *track;
	track_frame_t *frame;
	njbttaghdr_t header;
	unsigned char *tag;
	u_int32_t tagsize;
	int replace;

	if ((status = njb_usb_check(njb)) < 0)
		return NULL;

	/* check if track id has been given */
	if (*trackid > 0) {
		header.trackid = *trackid;
		header.size = fattr->track->header.size;	/* the old header size */
		dbg
		    ("update_track_tag: replacing, getting track id = %d, tag size = %d",
		     header.trackid, header.size);
		track = njb_usb_get_track_tag(njb, &header);
		if (track == NULL) {
			err
			    ("update_track_tag: can't get track tag for id = %d",
			     header.trackid);
			return NULL;
		}
		replace = 1;
	} else {
		track = track_new();
		dbg
		    ("update_track_tag: new track tag, track id unknown yet");
		if (track == NULL) {
			err("update_track_tag: no memory for track");
			return NULL;
		}
		replace = 0;
	}

	if (fattr->codec == NULL) {
		err("update_track_tag: fatal, no codec defined");
		return NULL;
	}
	frame = track_newframe(6, FR_CODEC, ID_DATA_ASCII,
			       strlen(fattr->codec) + 1, fattr->codec);
	if (frame == NULL) {
		err("update_track_tag: can't create codec frame");
		return NULL;
	}
	track_updateframe(track, frame);
	dbg("codec = \"%s\"", fattr->codec);

	if (fattr->title) {
		frame = track_newframe(6, FR_TITLE, ID_DATA_ASCII,
				       strlen(fattr->title) + 1,
				       fattr->title);
		if (frame == NULL) {
			err("update_track_tag: can't create title frame");
			return NULL;
		}
	}
	track_updateframe(track, frame);
	dbg("title = \"%s\"", fattr->title);

	if (fattr->artist) {
		frame = track_newframe(7, FR_ARTIST, ID_DATA_ASCII,
				       strlen(fattr->artist) + 1,
				       fattr->artist);
		if (frame == NULL) {
			err("update_track_tag: can't create artist frame");
			return NULL;
		}
	}
	track_updateframe(track, frame);
	dbg("artist = \"%s\"", fattr->artist);

	frame = track_newframe(10, FR_SIZE, ID_DATA_BIN,
			       sizeof(u_int32_t), &fattr->size);
	if (frame == NULL) {
		err("update_track_tag: can't create size frame");
		return NULL;
	}
	track_updateframe(track, frame);
	dbg("size = \"%d\" ", fattr->size);

	if (fattr->album) {
		frame = track_newframe(6, FR_ALBUM, ID_DATA_ASCII,
				       strlen(fattr->album) + 1,
				       fattr->album);
		if (frame == NULL) {
			err("update_track_tag: can't create album frame");
			return NULL;
		}
		track_updateframe(track, frame);
		dbg("album = \"%s\"", fattr->album);
	}

	if (fattr->genre) {
		if ((frame = track_newframe(6, FR_GENRE, ID_DATA_ASCII,
					    strlen(fattr->genre) + 1,
					    fattr->genre)) == NULL) {
			err("update_track_tag: can't create genre frame");
			return NULL;
		}
		track_updateframe(track, frame);
		dbg("genre = \"%s\"", fattr->genre);
	}


	if (fattr->length > 0) {
		if ((frame = track_newframe(7, FR_LENGTH, ID_DATA_BIN,
					    sizeof(u_int32_t),
					    &fattr->length)) == NULL) {
			err("update_track_tag: can't create length frame");
			return NULL;
		}
		track_updateframe(track, frame);
		dbg("length = \"%d\"", fattr->length);
	}

	if (fattr->tracknum > 0) {
		if ((frame = track_newframe(10, FR_TRACK, ID_DATA_BIN,
					    sizeof(u_int32_t),
					    &fattr->tracknum)) == NULL) {
			err
			    ("update_track_tag: can't create track num frame");
			return NULL;
		}
		track_updateframe(track, frame);
		dbg("tracknum = \"%d\"", fattr->tracknum);
	}

	/* Pack the tag, get the tag size */

	if ((tag = track_pack(track, &tagsize)) == NULL) {
		err("update_track_tag: can't pack track");
		return NULL;
	}
	dbg("update_track_tag: track packed, tag size = %d", tagsize);

	/* Check if replace? */

	if (replace) {
		/* replace tag track */
		header.size = tagsize;
		header.trackid = *trackid;
		dbg
		    ("update_track_tag: sending command 'replace track tag', tagsize = %d ",
		     tagsize);
		status = njb_usb_replace_track_tag(njb, tag, &header);
		if (status < 0) {
			err("update_track_tag: replace error, status = %d",
			    status);
			return NULL;
		}

	} else {
		/* Send track tag */
		header.size = tagsize;
		header.trackid = *trackid;
		dbg
		    ("update_track_tag: sending command 'send track tag', tagsize = %d ",
		     tagsize);
		status = njb_usb_new_track_tag(njb, tag, &header);
		if (status < 0) {
			err
			    ("update_track_tag: sending track tag error, status = %d",
			     status);
			return NULL;
		}

	}
	return track;
}


/**
 *  Get a playlist header from the NJB
 *  If there is not a next playlist header, the function returns <0
 *
 *  @param njb the njb to get the playlist from
 *  @param header the header to be filled with the result
 *  @param cmd can be NJB_GET_FIRST_PLAYLIST_HEADER or
 *           NJB_GET_NEXT_PLAYLIST_HEADER
 *  @return 0 if everything is ok or <0 if there was an error
 */
static int
njb_usb_get_playlist_header(struct nomad_usb_data *njb,
			    njbplhdr_t * header, int cmd)
{
	int status;
	unsigned char data[9];

	if ((status = njb_usb_check(njb)) < 0)
		return status;

	status = njb_usb_control(njb,
				 usb_rcvctrlpipe(njb->nomad_dev, 0),
				 cmd,
				 UT_READ | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, 0, 0, &data, 9, 3);
	if (status < 0) {
		err("get_playlist_header: usb control error");
		return -EIO;
	}

	if (data[0] == NJB_TRACK_NOT_FOUND) {
		return -ENOENT;
	} else if (data[0]) {
		err("get_playlist_header: bad NJB status = %d",
		    (int) data[0]);
		return -EAGAIN;
	}

	memcpy(&header->plid, &data[1], 4);
	memcpy(&header->size, &data[5], 4);

	return 0;
}

/**
 *  Get a playlist from the NJB.
 *  This command is used to retrieve a specific playlist
 *  from the NJB. The desired playlist is identified by the "playlist ID"
 *  that was returned from a previous "Get Playlist Header" request. The
 *  playlist header also indicates how large the playlist is, so that we
 *  can make appropriate storage arrangements.
 *  There is no data phase to the request. Instead, the NJB responds to the
 *  request by sending the playlist definition via the bulk data pipe.
 *
 *  @param njb the NJB to get the playlist from
 *  @param header a valid playlist header from a previous "Get Playlist Header"
 *  @return the playlist
 *
 */
static playlist_t *njb_usb_get_playlist(struct nomad_usb_data *njb,
					njbplhdr_t * header)
{
	int status;
	u_int16_t msw, lsw;
	void *data;
	playlist_t *playlist;
	unsigned char *dp;
	unsigned char dummy;
	int actual_length = 0;

	if ((status = njb_usb_check(njb)) < 0)
		return NULL;

	if ((data = kmalloc(header->size + 4, GFP_KERNEL)) == NULL) {
		err("couldn't kmalloc playlist data");
		return NULL;
	}

	msw = MSW(header->plid);
	lsw = LSW(header->plid);

	/*  send a get playlist command */
	status = njb_usb_control(njb,
				 usb_sndctrlpipe(njb->nomad_dev, 0),
				 NJB_CMD_GET_PLAYLIST,
				 UT_WRITE | USB_TYPE_VENDOR |
				 USB_RECIP_OTHER, msw, lsw, &dummy, 0, 3);
	if (status < 0) {
		err("njb_usb_get_playlist: njb control error");
		return NULL;
	}

	/*  receive usb bulk read */
	status = usb_bulk_msg(njb->nomad_dev,
			      usb_rcvbulkpipe(njb->nomad_dev,
					      NJB_ENDPOINT_BULK), data,
			      header->size + 5, &actual_length,
			      USB_BULK_TIMEOUT);

	if (status == -1) {
		err("njb_usb_get_playlist: bulk read error");
		return NULL;
	} else if (actual_length < (header->size + 5)) {
		err("njb_usb_get_playlist: short read, actual_length = %d",
		    actual_length);
		return NULL;
	}

	/* the playlist is in NJB format, so we convert it to our own */
	dp = (unsigned char *) data;
	playlist = playlist_unpack(&dp[5], header->size);
	if (playlist == NULL)
		return NULL;

	return playlist;
}


#ifdef DEBUG
void dump_data(unsigned char *data, int count)
{
	unsigned char *ptr;
	unsigned char buf[50];
	unsigned char chars[18];
	int x, y;

	ptr = data;
	for (y = 0; y < count; y += 16) {
		for (x = 0; x < 16 && x < count; x++) {
			sprintf(&buf[x * 3], "%02x ", *ptr);
			chars[x] = isalnum(*ptr) ? *ptr : '.';
			ptr++;
		}
		chars[16] = '\0';
		info("dump: %s %s", buf, chars);
	}
}
#endif


EXPORT_SYMBOL(njb_usb_connect);
EXPORT_SYMBOL(njb_usb_disconnect);
EXPORT_SYMBOL(njb_usb_open);
EXPORT_SYMBOL(njb_usb_close);
EXPORT_SYMBOL(njb_usb_read);
EXPORT_SYMBOL(njb_usb_write);
EXPORT_SYMBOL(njb_usb_rename);
EXPORT_SYMBOL(njb_usb_delete);
EXPORT_SYMBOL(njb_usb_create);
EXPORT_SYMBOL(njb_usb_capture);
EXPORT_SYMBOL(njb_usb_release);
EXPORT_SYMBOL(njb_usb_get_track_tag);
EXPORT_SYMBOL(njb_usb_get_first_track_tag_header);
EXPORT_SYMBOL(njb_usb_get_next_track_tag_header);
