
/*
 * Driver for USB Creative MP3 Portable Player
 *
 * Karl Backstrom <backstrom@kde.org>
 *
 * based on nomadII-driver.c by Gabriel Gomiz (gomita@bblanca.com.ar)
 *
 * based on rio500.c by Cesar Miquel (miquel@df.uba.ar)
 *
 * based on hp_scanner.c by David E. Nelson (dnelson@jump.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Based upon mouse.c (Brad Keryan) and printer.c (Michael Gee).
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/malloc.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#if defined(CONFIG_DEVFS_FS)
#include <linux/devfs_fs_kernel.h>
#endif

#if !defined (DEBUG) && defined (CONFIG_USB_DEBUG)
#   define DEBUG
#endif
#include <linux/usb.h>

#include "nomad_usb.h"

#if defined(CONFIG_DEVFS_FS)
devfs_handle_t nomad_dev;
extern devfs_handle_t usb_devfs_handle;
extern struct file_operations usb_fops;
#endif

static int deviceregistered[MAX_NO_OF_DEVICES];

static struct nomad_usb_data nomad_instance;

static int open_nomad(struct inode *inode, struct file *file)
{
	struct nomad_usb_data *nomad = &nomad_instance;

	lock_kernel();

	if (nomad->isopen || !nomad->present) {
		unlock_kernel();
		return -EBUSY;
	}
	nomad->isopen = 1;

	/* Initialize queue of waiting process */
	init_waitqueue_head(&nomad->wait_q);

	MOD_INC_USE_COUNT;

	unlock_kernel();

	info("Nomad device opened.");

	return 0;
}

static int close_nomad(struct inode *inode, struct file *file)
{
	struct nomad_usb_data *nomad = &nomad_instance;

	nomad->isopen = 0;

	MOD_DEC_USE_COUNT;

	info("Nomad deviceclosed.");
	return 0;
}

static int
ioctl_nomad(struct inode *inode, struct file *file, unsigned int cmd,
	    unsigned long arg)
{
	struct Nomad_cmd nomad_cmd;
	struct nomad_usb_data *nomad = &nomad_instance;
	void *data;
	unsigned char *buffer;
	int result, requesttype;
	int retries;

	result = -1;

	/* Sanity check to make sure Nomad is connected, powered, etc */
	if (nomad == NULL ||
	    nomad->present == 0 ||
	    nomad->nomad_dev == NULL || nomad->isopen == 0)
		return -EIO;

	switch (cmd) {
	case NOMAD_RECV_COMMAND:
		data = (void *) arg;
		if (data == NULL)
			break;
		if (copy_from_user
		    (&nomad_cmd, data,
		     sizeof(struct Nomad_cmd))) return -EFAULT;

		if (nomad_cmd.length > PAGE_SIZE)
			return -EINVAL;

		buffer = (unsigned char *) __get_free_page(GFP_KERNEL);

		if (buffer == NULL)
			return -ENOMEM;

		if (copy_from_user
		    (buffer, nomad_cmd.buffer,
		     nomad_cmd.length)) return -EFAULT;

		requesttype =
		    nomad_cmd.
		    requesttype | USB_DIR_IN | USB_TYPE_VENDOR |
		    USB_RECIP_DEVICE;

		info
		    ("sending command: reqtype=%0x req=%0x value=%0x index=%0x len=%0x",
		     requesttype, nomad_cmd.request, nomad_cmd.value,
		     nomad_cmd.index, nomad_cmd.length);

		/* Send nomad control message */
		retries = 3;
		while (retries) {

			result = usb_control_msg(nomad->nomad_dev,
						 usb_rcvctrlpipe(nomad->
								 nomad_dev,
								 ENDPOINT_CONTROL),
						 nomad_cmd.request,
						 requesttype,
						 nomad_cmd.value,
						 nomad_cmd.index, buffer,
						 nomad_cmd.length,
						 nomad_cmd.timeout);

			if (result == -ETIMEDOUT)
				retries--;

			else if (result < 0) {

				err("Error executing ioctl. code = %d",
				    le32_to_cpu(result));
				retries = 0;

			} else {

				info
				    ("Executed ioctl. Result = %d (data=%04x)",
				     le32_to_cpu(result),
				     le32_to_cpu(*((long *) buffer)));
				if (copy_to_user
				    (nomad_cmd.buffer, buffer,
				     nomad_cmd.length)) return -EFAULT;
				retries = 0;

			}

		}
		free_page((unsigned long) buffer);
		break;

	case NOMAD_SEND_COMMAND:

		data = (void *) arg;

		if (data == NULL)
			break;

		if (copy_from_user
		    (&nomad_cmd, data,
		     sizeof(struct Nomad_cmd))) return -EFAULT;

		if (nomad_cmd.length > PAGE_SIZE)
			return -EINVAL;

		buffer = (unsigned char *) __get_free_page(GFP_KERNEL);

		if (buffer == NULL)
			return -ENOMEM;

		if (copy_from_user
		    (buffer, nomad_cmd.buffer,
		     nomad_cmd.length)) return -EFAULT;

		requesttype =
		    nomad_cmd.
		    requesttype | USB_DIR_OUT | USB_TYPE_VENDOR |
		    USB_RECIP_DEVICE;

		info
		    ("sending command: reqtype=%0x req=%0x value=%0x index=%0x len=%0x",
		     requesttype, nomad_cmd.request, nomad_cmd.value,
		     nomad_cmd.index, nomad_cmd.length);

		/* Send nomad control message */
		retries = 3;
		while (retries) {

			result = usb_control_msg(nomad->nomad_dev,
						 usb_sndctrlpipe(nomad->
								 nomad_dev,
								 ENDPOINT_CONTROL),
						 nomad_cmd.request,
						 requesttype,
						 nomad_cmd.value,
						 nomad_cmd.index, buffer,
						 nomad_cmd.length,
						 nomad_cmd.timeout);

			if (result == -ETIMEDOUT)
				retries--;

			else if (result < 0) {

				err("Error executing ioctl. code = %d",
				    le32_to_cpu(result));
				retries = 0;

			} else {

				info("Executed ioctl. Result = %d",
				     le32_to_cpu(result));
				retries = 0;

			}

		}

		free_page((unsigned long) buffer);
		break;

	case NOMAD_RESET_PIPES:

		if (usb_clear_halt(nomad->nomad_dev, ENDPOINT_BULK_READ)) {
			info("reset pipes request failed.");
			return -EPIPE;
		}
		if (usb_clear_halt(nomad->nomad_dev, ENDPOINT_BULK_WRITE)) {
			info("reset pipes request failed.");
			return -EPIPE;
		}

		return 0;

	default:
		return -ENOIOCTLCMD;
		break;
	}

	return result;
}

static ssize_t
write_nomad(struct file *file, const char *buffer, size_t count,
	    loff_t * ppos)
{
	struct nomad_usb_data *nomad = &nomad_instance;

	unsigned long copy_size;
	unsigned long bytes_written = 0;
	unsigned int partial;

	int result = 0;
	int maxretry;
	int errn = 0;

	/* Sanity check to make sure nomad is connected, powered, etc */
	if (nomad == NULL ||
	    nomad->present == 0 ||
	    nomad->nomad_dev == NULL || nomad->isopen == 0)
		return -1;

	down(&(nomad->lock));

	do {
		unsigned long thistime;
		char *obuf = nomad->obuf;

		thistime = copy_size =
		    (count >= OBUF_SIZE) ? OBUF_SIZE : count;

		if (copy_from_user(nomad->obuf, buffer, copy_size)) {
			errn = -EFAULT;
			goto error;
		}

		maxretry = 5;
		//pipe = usb_sndbulkpipe(nomad->nomad_dev,ENDPOINT_BULK_WRITE);
		while (thistime) {
			if (!nomad->nomad_dev) {
				errn = -ENODEV;
				goto error;
			}

			if (signal_pending(current)) {
				up(&(nomad->lock));
				return bytes_written ? bytes_written :
				    -EINTR;
			}

			result = usb_bulk_msg(nomad->nomad_dev,
					      usb_sndbulkpipe(nomad->
							      nomad_dev,
							      2), obuf,
					      thistime, &partial, 5 * HZ);
			info
			    ("write stats: result:%d thistime:%lu partial:%u",
			     result, thistime, partial);

			if (result == USB_ST_TIMEOUT) {	/* NAK - so hold for a while */
				if (!maxretry--) {
					errn = -ETIME;
					goto error;
				}
				interruptible_sleep_on_timeout(&nomad->
							       wait_q,
							       NAK_TIMEOUT);
				continue;
				/*} else if (result == -EPIPE) {
				   warn("CLEAR FEATURE requested to remove STALL condition.");
				   if (usb_clear_halt(nomad->nomad_dev,usb_pipeendpoint(pipe)))
				   err("request failed.");
				   break; */

			} else if (!result && partial) {
				obuf += partial;
				thistime -= partial;
			} else
				break;
		};

		if (result) {
			err("write_nomad: Whoops - %x", result);
			errn = -EIO;
			goto error;
		}
		bytes_written += copy_size;
		count -= copy_size;
		buffer += copy_size;
	} while (count > 0);

	up(&(nomad->lock));

	return bytes_written ? bytes_written : -EIO;

      error:
	up(&(nomad->lock));
	return errn;
}


static ssize_t
read_nomad(struct file *file, char *buffer, size_t count, loff_t * ppos)
{
	struct nomad_usb_data *nomad = &nomad_instance;
	ssize_t read_count;
	unsigned int partial;
	int this_read;
	int result;
	int maxretry = 10;
	char *ibuf = nomad->ibuf;

	/* Sanity check to make sure Nomad is connected, powered, etc */
	if (nomad == NULL ||
	    nomad->present == 0 ||
	    nomad->nomad_dev == NULL || nomad->isopen == 0)
		return -1;

	/* Do the reading... */
	read_count = 0;

	down(&(nomad->lock));
	//pipe = usb_rcvbulkpipe(nomad->nomad_dev,ENDPOINT_BULK_READ);

	while (count > 0) {
		if (signal_pending(current)) {
			up(&(nomad->lock));
			return read_count ? read_count : -EINTR;
		}
		/* Sanity check */
		if (!nomad->nomad_dev) {
			up(&(nomad->lock));
			return -ENODEV;
		}

		this_read = (count >= IBUF_SIZE) ? IBUF_SIZE : count;

		result = usb_bulk_msg(nomad->nomad_dev,
				      usb_rcvbulkpipe(nomad->nomad_dev, 1),
				      ibuf, this_read, &partial, 5 * HZ);

		info("read stats: result:%d this_read:%u partial:%u",
		     result, this_read, partial);

		if (partial) {
			count = this_read = partial;
		} else if (result == USB_ST_TIMEOUT) {
			if (!maxretry--) {
				up(&(nomad->lock));
				err("read_nomad: maxretry timeout");
				return -ETIME;
			}
			interruptible_sleep_on_timeout(&nomad->wait_q,
						       NAK_TIMEOUT);
			continue;

			/*} else if (result == -EPIPE) {
			   warn("CLEAR FEATURE requested to remove STALL condition.");
			   if (usb_clear_halt(nomad->nomad_dev,usb_pipeendpoint(pipe)))
			   err("request failed.");
			   break; */

		} else if (result != USB_ST_DATAUNDERRUN) {
			up(&(nomad->lock));
			err
			    ("read_nomad: whoops - result:%u partial:%u this_read:%u",
			     result, partial, this_read);
			return -EIO;
		} else {
			unlock_kernel();
			return (0);

		}		/* end if (partial) */

		if (this_read) {
			if (copy_to_user(buffer, ibuf, this_read)) {
				up(&(nomad->lock));
				return -EFAULT;
			}
			count -= this_read;
			read_count += this_read;
			buffer += this_read;
		}
	}
	up(&(nomad->lock));
	return read_count;
}

static struct file_operations usb_nomad_fops = {
	read:read_nomad,
	write:write_nomad,
	ioctl:ioctl_nomad,
	open:open_nomad,
	release:close_nomad,
};

struct nomad_info nomad_info[] = {
	{"Nomad Jukebox"},
	{"Nomad II"},
	{"Nomad II MG"},
	{NULL}
};

struct usb_device_id nomad_table[] = {
      {USB_DEVICE(0x0471, 0x0222), driver_info:0},
				/* Nomad Jukebox */
      {USB_DEVICE(0x041e, 0x1002), driver_info:1},
				/* NOMAD II      */
      {USB_DEVICE(0x041e, 0x4004), driver_info:2},
				/* Nomad II MG   */
	{}			/* Terminating entry */
};

MODULE_DEVICE_TABLE(usb, nomad_table);

static void *probe_nomad(struct usb_device *dev, unsigned int ifnum,
			 const struct usb_device_id *id)
{
	int i;
	struct nomad_usb_data *nomad = &nomad_instance;

#if defined(CONFIG_DEVFS_FS)
	char *devfs_name = "nomad0";
#endif

	nomad->device_counter = -1;
	for (i = 0; i < MAX_NO_OF_DEVICES; i++) {
		if (deviceregistered[i] == 0) {
			nomad->device_counter = i;
			deviceregistered[i] = 1;
			i = MAX_NO_OF_DEVICES;	/* break; */
		}
	}
	if (nomad->device_counter >= 0) {

		nomad->info = nomad_info + id->driver_info;	/* device name */
		nomad->present = 1;
		nomad->nomad_dev = dev;
		nomad->captured = 0;
		nomad->session_updated = 0;
		init_MUTEX(&(nomad->lock));
		/* nomad->nomad_dev.name = nomad->name; */

		info
		    ("probe_nomad: USB '%s' found at address %d, device counter = %d",
		     nomad->info->name, dev->devnum,
		     nomad->device_counter);

		if (!
		    (nomad->obuf =
		     (char *) kmalloc(OBUF_SIZE, GFP_KERNEL))) {
			err
			    ("probe_nomad: Not enough memory for the output buffer");
			return NULL;
		}
		dbg("probe_nomad: obuf address:%p", nomad->obuf);

		if (!
		    (nomad->ibuf =
		     (char *) kmalloc(IBUF_SIZE, GFP_KERNEL))) {
			err
			    ("probe_nomad: Not enough memory for the input buffer");
			kfree(nomad->obuf);
			return NULL;
		}
		dbg("probe_nomad: ibuf address:%p", nomad->ibuf);

#if defined(CONFIG_DEVFS_FS)
		sprintf(devfs_name, "nomad%d", nomad->device_counter);

		nomad_dev = devfs_register(usb_devfs_handle,
					   devfs_name, DEVFS_FL_NONE,
					   USB_MAJOR,
					   NOMAD_MINOR +
					   nomad->device_counter,
					   S_IFCHR | S_IRUGO | S_IWUSR,
					   &usb_nomad_fops, NULL);
		info
		    ("probe_nomad: devfs_register executed, nomad_dev = %d",
		     nomad_dev);
#endif
		return nomad;
	}
	err("probe_nomad: Maximum number of devices reached");
	return NULL;
}

static void disconnect_nomad(struct usb_device *dev, void *ptr)
{
	struct nomad_usb_data *nomad = (struct nomad_usb_data *) ptr;

	if (nomad->isopen) {
		nomad->isopen = 0;
		/* better let it finish - the release will do whats needed */
		nomad->nomad_dev = NULL;
		return;

	}
	kfree(nomad->ibuf);
	kfree(nomad->obuf);

	info("USB device '%s' disconnected.", nomad->info->name);

	nomad->present = 0;
	deviceregistered[nomad->device_counter] = 0;
#if defined(CONFIG_DEVFS_FS)
	devfs_unregister(nomad_dev);
#endif
}

static struct usb_driver nomad_driver = {
	name:"nomad",
	probe:probe_nomad,
	disconnect:disconnect_nomad,
	fops:&usb_nomad_fops,
	minor:NOMAD_MINOR,
	id_table:nomad_table,
};

int usb_nomad_init(void)
{
	int i;
	if (usb_register(&nomad_driver) < 0)
		return -1;

	for (i = 0; i < MAX_NO_OF_DEVICES; i++)
		deviceregistered[i] = 0;

	info("USB Creative Nomad II/Jukebox support registered.");
	return 0;
}


void usb_nomad_cleanup(void)
{

	struct nomad_usb_data *nomad = &nomad_instance;

	nomad->present = 0;
#if defined(CONFIG_DEVFS_FS)
	devfs_unregister(nomad_dev);
#endif
	usb_deregister(&nomad_driver);
}

EXPORT_SYMBOL(usb_nomad_init);
EXPORT_SYMBOL(usb_nomad_cleanup);

/* -----------original module description------------------------------- */

/* overridden in inode.c */
#if 0
MODULE_AUTHOR("Karl Backstrom <backstrom@kde.org>");
MODULE_DESCRIPTION("USB Creative Nomad Driver for Linux (c)2000");

module_init(usb_nomad_init);
module_exit(usb_nomad_cleanup);
#endif
/* --------------------------------------------------------------------- */
