/*
 * Nomad Jukebox driver for Linux
 *
 * Jrg Prante <joergprante@gmx.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/slab.h>

#include "njbfs.h"


static track_t *track_new(void);
static track_t *track_unpack(void *data, size_t nbytes);
static unsigned char *track_pack(track_t * track, u_int32_t * size);
static void track_destroy(track_t * track);
static void track_reset_getframe(track_t * track);
static track_frame_t *track_getframe(track_t * track);
static track_frame_t *track_findframe(track_t * track, const char *label);
static track_frame_t *track_newframe(u_int16_t labelsz, const void *label,
				     u_int16_t type, u_int16_t datasz,
				     const void *data);
static void track_addframe(track_t * track, track_frame_t * frame);
static void track_replaceframe(track_t * track, track_frame_t * oldframe,
			       track_frame_t * newframe);
static void track_frame_destroy(track_frame_t * frame);
static void track_updateframe(track_t * track, track_frame_t * frame);


static track_t *track_new(void)
{
	track_t *track;

	track = (track_t *) kmalloc(sizeof(track_t), GFP_KERNEL);
	if (track == NULL) {
		return NULL;
	}

	memset(track, 0, sizeof(track_t));

	return track;
}

static track_t *track_unpack(void *data, size_t nbytes)
{
	unsigned char *dp = (unsigned char *) data;
	size_t index;
	track_t *track;
	u_int16_t i, nframes;

	track = track_new();
	if (track == NULL)
		return NULL;

	memcpy(&nframes, dp, 2);

	index = 2;
	for (i = 0; i < nframes; i++) {
		u_int16_t lsize, vsize, type;
		track_frame_t *frame;

		memcpy(&type, &dp[index], 2);
		index += 2;
		memcpy(&lsize, &dp[index], 2);
		index += 2;
		memcpy(&vsize, &dp[index], 2);
		index += 4;

		frame =
		    track_newframe(lsize, &dp[index], type, vsize,
				   &dp[index + lsize]);
		if (frame == NULL) {
			track_destroy(track);
			return NULL;
		}
		track_addframe(track, frame);
		index += (lsize + vsize);

		if (index > nbytes) {
			track_destroy(track);
			return NULL;
		}
	}

	return track;
}

static unsigned char *track_pack(track_t * track, u_int32_t * tagsize)
{
	track_frame_t *frame;
	unsigned char *data;
	size_t index;
	u_int32_t i32;

	*tagsize = 0;

	if (!(track->nframes)) {
		return NULL;
	}

	track_reset_getframe(track);
	while ((frame = track_getframe(track)) != NULL) {
		*tagsize += 8 + frame->labelsz + frame->datasz;
	}

	if (!(*tagsize)) {
		return NULL;
	}

	*tagsize += 2;
	data = (unsigned char *) kmalloc(*tagsize, GFP_KERNEL);
	if (data == NULL) {
		return NULL;
	}

	i32 = track->nframes;
	memcpy(data, &i32, 2);

	index = 2;
	track_reset_getframe(track);
	while ((frame = track_getframe(track)) != NULL) {
		i32 = frame->type;
		memcpy(&data[index], &i32, 2);
		index += 2;

		i32 = frame->labelsz;
		memcpy(&data[index], &i32, 2);
		index += 2;

		i32 = frame->datasz;
		memcpy(&data[index], &i32, 2);
		index += 2;

		memset(&data[index], 0, 2);
		index += 2;

		memcpy(&data[index], frame->label, frame->labelsz);
		index += frame->labelsz;

		memcpy(&data[index], frame->data, frame->datasz);
		index += frame->datasz;
	}
	return data;
}




static void track_addframe(track_t * track, track_frame_t * frame)
{
	if (track->nframes) {
		track->last->next = frame;
		track->last = frame;
	} else {
		track->first = track->last = frame;
		track->cur = track->first;
	}

	frame->next = NULL;
	track->nframes++;
}

static void
track_replaceframe(track_t * track, track_frame_t * oldframe,
		   track_frame_t * newframe)
{
	track_frame_t *frame;

	track_reset_getframe(track);
	while ((frame = track_getframe(track)) != NULL) {
		if (!strcmp(frame->label, oldframe->label)) {
			track_frame_destroy(oldframe);
			track_addframe(track, newframe);
			return;
		}
	}

}

static void track_updateframe(track_t * track, track_frame_t * newframe)
{
	track_frame_t *frame;

	track_reset_getframe(track);
	while ((frame = track_getframe(track)) != NULL) {
		if (!strcmp(frame->label, newframe->label)) {
			track_frame_destroy(frame);
			track_addframe(track, newframe);
			return;
		}
	}
	track_addframe(track, newframe);

}

static void track_destroy(track_t * track)
{
	track_frame_t *frame;

	track_reset_getframe(track);
	while ((frame = track_getframe(track)) != NULL) {
		track_frame_destroy(frame);
	}

	kfree(track);
}

static void track_reset_getframe(track_t * track)
{
	track->cur = track->first;
}

static track_frame_t *track_getframe(track_t * track)
{
	track_frame_t *t;

	if (track->cur == NULL)
		return NULL;
	t = track->cur;
	track->cur = track->cur->next;

	return t;
}

static track_frame_t *track_findframe(track_t * track, const char *label)
{
	track_frame_t *frame;

	track_reset_getframe(track);
	while ((frame = track_getframe(track)) != NULL) {
		if (!strcmp(frame->label, label))
			return frame;
	}

	return NULL;
}


static track_frame_t *track_newframe(u_int16_t labelsz, const void *label,
				     u_int16_t type, u_int16_t datasz,
				     const void *data)
{
	track_frame_t *frame;
	u_int16_t olddatasz;

	frame =
	    (track_frame_t *) kmalloc(sizeof(track_frame_t), GFP_KERNEL);
	if (frame == NULL) {
		return NULL;
	}

	/* align to long words, pad with null values */
	while ((labelsz & 0x03) != 0) {
		labelsz++;
	}

	olddatasz = datasz;
	while ((datasz & 0x03) != 0) {
		datasz++;
	}

	frame->label = kmalloc(labelsz, GFP_KERNEL);
	frame->data = kmalloc(datasz, GFP_KERNEL);

	if (frame->label == NULL || frame->data == NULL) {
		track_frame_destroy(frame);
		return NULL;
	}
	memset(frame->label, 0, labelsz);
	memset(frame->data, 0, datasz);

	frame->labelsz = labelsz;
	frame->datasz = datasz;
	frame->type = type;

	/* frame label is always a string */
	strcpy(frame->label, label);
	/* data can be text or binary */
	memcpy(frame->data, data, olddatasz);

	return frame;
}



static void track_frame_destroy(track_frame_t * frame)
{
	if (frame->label != NULL)
		kfree(frame->label);
	if (frame->data != NULL)
		kfree(frame->data);
	kfree(frame);
}

EXPORT_SYMBOL(track_unpack);
EXPORT_SYMBOL(track_pack);
EXPORT_SYMBOL(track_findframe);
EXPORT_SYMBOL(track_newframe);
EXPORT_SYMBOL(track_addframe);
EXPORT_SYMBOL(track_replaceframe);
EXPORT_SYMBOL(track_updateframe);
